////////////////////////////////////////////////////////////////////////////////
//
//	Russell Osterlund, May 2001
//
//	Pseudocode for:
//		LdrpClearLoadInProgress
//		LdrpLoadDll
//
////////////////////////////////////////////////////////////////////////////////

#include "LdrpLoadDll.h"

////////////////////////////////////////////////////////////////////////////////
// This routine is normally called by LdrpRunInitializeRoutines.
ULONG
	LdrpClearLoadInProgress
		()
{
	ULONG Count = 0;
	PHEADERLINK pInitOrderHeaderLink =
		((PMODULELISTHEADER) (PEB + 0xc))->InitOrder;
	PMODULEITEM pModuleItem = 
		(PMODULEITEM) pInitOrderHeaderLink->First;
	while (pModuleItem != pInitOrderHeaderLink)
	{
		if (pModuleItem->ModuleFlags & LOAD_IN_PROGRESS)
			pModuleItem->ModuleFlags &= ~LOAD_IN_PROGRESS;
		if (pModuleItem->ModuleFlags & ~ENTRY_PROCESSED)
			if (pModuleItem->MemOrder.Prev != NULL)
				Count++;
		pModuleItem = pModuleItem->LoadOrder.Next;
	}
	return Count;
 // LdrpClearLoadInProgress

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpLoadDll
		(LPCWSTR SearchPath,
		PULONG ResolveDllReferences,
		PUNICODE_STRING DllName,
		PULONG ImageBase,			// OUT
		ULONG NormalProcessing)		// 0 - forwarded processing
									// 1 - normal processing
{
	LONG LoadingPhase;						// ebp-0x4
	ULONG Unknown008;						// ebp-0x8
	ULONG StackFramePtr;					// ebp-0x10
	LPCWSTR DllNameBuffer;					// ebp-0x1C
	UNICODE_STRING WorkUnicodeString;		// ebp-0x24
	PMODULEITEM pModuleItem;				// ebp-0x2C
	LPCWSTR DotPosition;					// ebp-0x34
	NTSTATUS LdrpLoadDllResult;				// ebp-0x38
	ULONG Unknown03C;						// ebp-0x3C
	ULONG ReturnCode;						// ebp-0x40
	LPCWSTR WorkName;						// ebp-0x254
	PMODULELISTLINKS pInitOrderHeaderLink;	// ebp-0x264

	// This routine is wrapped by a __try/__except block.
	ReturnCode = 0;
	// Try to be optimistic!
	LdrpLoadDllResult = STATUS_SUCCESS;
	LoadingPhase = 0;
	if (LdrpInLdrInit)
		RltEnterCriticalSection
			(PEB->CriticalSectionArea);		// +0xA0
	DllNameBuffer = DllName->Buffer;
	DotPosition = 
		wcsrchr
			(DllName->Buffer,
			'.');
	if (DllName->Length > MAX_DLL_NAME_LENGTH)
	{
		_local_unwind2
			(StackFramePtr,
			-1);
		// Set current frame pointer to StackFramePtr
		return (STATUS_NAME_TOO_LONG);
	}
	memmove
		(WorkName,
		DllName->Buffer,
		DllName->Length);
	// If the DllName passed to us does not include the .dll extension, add it.
	if ((DotPosition)
			&&
		(DotPosition != '\\'))
		;
	else
	{
		if (DllName->Length + 0x10 > MAX_DLL_NAME_LENGTH)
		{
			_local_unwind2
				(StackFramePtr,
				-1);
			// Set current frame pointer to StackFramePtr
			return (STATUS_NAME_TOO_LONG);
		}
		memmove
			(WorkName[DllName->Length],
			_T(".dll"),
			0x10);
	}
	if (ShowSnaps)
		if (SearchPath)
			DbgPrint
				("LdrLoadDll, loading %ws from %ws",
				DllName,
				SearchPath);
		else
			DbgPrint
				("LdrLoadDll, loading %ws from %ws",
				DllName,
				0x77F9B4BC);
	RtlInitUnicodeString
		(WorkUnicodeString,
		WorkName);
	WorkUnicodeString.MaximumLength = MAX_DLL_NAME_LENGTH;
	if (!LdrpCheckForLoadedDll
			(SearchPath,
			WorkUnicodeString,
			0,
			&ReturnCode,
			pModuleItem))
	{
		// The DLL has not been loaded; there is more work to do.
		LdrpLoadDllResult =
			LdrpMapDll
				(SearchPath,
				WorkName,				
				ResolveDllReferences,
				0,						// dynamic link
				&ReturnCode,
				pModuleItem);
		if (LdrpLoadDllResult != STATUS_SUCCESS)
		{
			_local_unwind2
				(StackFramePtr,
				-1);
			// Set current frame pointer to StackFramePtr
			return LdrpLoadDllResult;
		}
		if ((NULL != ResolveDllReferences)
				&&
			(*ResolveDllReferences & 0x02))	// i.e., do NOT resolve references
		{
			pModuleItem->EntryPoint = 0;
			pModuleItem->ModuleFlags &= ~IMAGE_DLL;
		}
		if (pModuleItem->ModuleFlags & IMAGE_DLL)
		{
			LoadingPhase = 1;
			LdrpLoadDllResult =
				LdrpWalkImportDescriptor
					(SearchPath,
					pModuleItem);
			LoadingPhase = 0;
			if (-1 != pModuleItem->LoadCount)
				pModuleItem->LoadCount++;
			LdrpUpdateLoadCount
				(pModuleItem,
				1);
			if (LdrpLoadDllResult < STATUS_SUCCESS)
			{
				// Something went wrong -- back out dll load.
				// Undo all of the linked-list processing.
				LdrpClearLoadInProgress();
				LdrpUnloadDll
					(pModuleItem);
				_local_unwind2
					(StackFramePtr,
					-1);
				// Set current frame pointer to StackFramePtr
				return LdrpLoadDllResult;
			}
		}
		else
		{
			if (-1 != pModuleItem->LoadCount)
				pModuleItem->LoadCount++;
		}
		pInitOrderHeaderLink =
			((PMODULELISTHEADER) (PEB + 0xc))->InitOrder;
		// Insert pModuleItem into the ModuleList's InitOrder chain (at the end)
		// and update the links in pModuleItem->InitOrder.
	}
	if ((!NormalProcessing)	// i.e., forwarded dll processing
			||
		(!LdrpLdrDatabaseIsSetup))
		LdrpLoadDllResult = STATUS_SUCCESS;
	else
	{
		LoadingPhase = 2;
		LdrpLoadDllResult =
			LdrpRunInitializeRoutines
				(0);
		if (STATUS_SUCCESS > LdrpLoadDllResult)
			LdrUnloadDll
				(pModuleItem->ImageBase);
		LoadingPhase = 0;
	}
	LoadingPhase |= -1;
	RltLeaveCriticalSection
		(PEB->CriticalSectionArea);		// +0xA0
	if (pModuleItem->ImageBase)
		*ImageBase = pModuleItem->ImageBase;
	else
		*ImageBase = 0;
	return LdrpLoadDllResult;
} // LdrpLoadDll