////////////////////////////////////////////////////////////////////////////////
//
//	Russell Osterlund, May 2001
//
//	Pseudocode for:
//		LdrpCheckForKnownDll
//		LdrpResolveDllName
//		LdrpCreateDllSection
//		LdrpAllocateDataTableEntry
//		LdrpFetchAddressOfEntryPoint
//		LdrpInsertMemoryTableEntry
//		LdrpMapDll
//
////////////////////////////////////////////////////////////////////////////////
#include <winnt.h>
#include <ntddk.h>
#include <ntdef.h>
#include "LdrpLoadDll.h"

////////////////////////////////////////////////////////////////////////////////
HANDLE
	LdrpCheckForKnownDll
		(LPCWSTR DllName,
		PUNICODE_STRING PathFileName,
		PUNICODE_STRING FileName)
{
	HANDLE KnownDll;					// ebp-0x4
	UNICODE_STRING UnicodeDllName;		// ebp-0xc
	OBJECT_ATTRIBUTES ObjectAttributes; // ebp-0x24

	KnownDll &= 0;
	RtlInitUnicodeString
		(&UnicodeDllName,
		DllName);
	FileName->Length = UnicodeDllName.Length;
	FileName->MaximumLength = UnicodeDllName.MaximumLength;
	FileName->Buffer =
		RtlAllocateHeap
			(GetProcessHeap(),
			MEM_MAPPED + NtdllBaseTag,
			UnicodeDllName.MaximumLength);
	if (NULL == FileName->Buffer)
		return 0;
	memmove
		(FileName->Buffer,
		UnicodeDllName.Buffer,
		UnicodeDllName.MaximumLength);
	PathFileName->Length += LdrpKnownDllPath.Length;
	PathFileName->MaximumLength += LdrpKnownDllPath.MaximumLength;
	PathFileName->Buffer = 
		RtlAllocateHeap
			(GetProcessHeap(),
			MEM_MAPPED + NtdllBaseTag,
			PathFileName->MaximumLength);
	if (NULL == PathFileName->Buffer)
	{
		RtlFreeHeap
			(GetProcessHeap(),
			FileName->Buffer,
			FileName->MaximumLength);
		return 0;
	}
	memmove
		(PathFileName->Buffer,
		LdrpKnownDllPath.Buffer,
		LdrpKnownDllPath.Length);
	PathFileName->Buffer[LdrpKnownDllPath.Length] = '\\';
	// Concatenate the unicode DllName to LdrpKnownDllPath.
	memmove
		(PathFileName->Buffer[LdrpKnownDllPath.Length + sizeof (WCHAR)],
		UnicodeDllName.Buffer,
		UnicodeDllName.Length);
	ObjectAttributes.RootDirectory = LdrpKnownDllObjectDirectory;
	ObjectAttributes.ObjectName = UnicodeDllName.Buffer;
	ObjectAttributes.Length = sizeof (OBJECT_ATTRIBUTES);
	ObjectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
	ObjectAttributes.SecurityDescriptor = NULL;
	ObjectAttributes.SecurityQualityOfService = NULL;
	KnownDll =
		NtOpenSection
			(&KnownDll,
			SECTION_MAP_WRITE | SECTION_MAP_READ | SECTION_MAP_EXECUTE,
			&ObjectAttributes);
	if (0 < KnownDll)
	{
		RtlFreeHeap
			(GetProcessHeap(),
			FileName->Buffer,
			FileName->MaximumLength);
		RtlFreeHeap
			(GetProcessHeap(),
			PathFileName->Buffer,
			PathFileName->MaximumLength);
	}
	return KnownDll;
} // LdrpCheckForKnownDll

////////////////////////////////////////////////////////////////////////////////
UCHAR
	LdrpResolveDllName
		(LPCWSTR SearchPath,
		LPCWSTR DllName,
		PUNICODE_STRING PathFileName,
		PUNICODE_STRING FileName,
		ULONG ReturnCode)
{
	ReturnCode = 0;
	PathFileName->Buffer =
		RtlAllocateHeap
			(GetProcessHeap(),
			(MEM_TOP_DOWN | MEM_RESET) + NtdllBaseTag,
			MAX_DLL_NAME_LENGTH);
	if (NULL == PathFileName->Buffer)
		return 0;
	if (NULL == SearchPath)
		// LdrpDefaultPath contains pointer to the process' search path.
		SearchPath = ((PUNICODE_STRING ) LdrpDefaultPath)->Buffer;	
	ULONG DosSearchPathResult =
		RtlDosSearchPath_U
			(SearchPath,
			DllName,
			0,
			MAX_DLL_NAME_LENGTH - sizeof (WCHAR),
			PathFileName->Buffer, 					// output
			&FileName->Buffer);						// output
	if ((DosSearchPathResult == 0)
			||
		(DosSearchPathResult > MAX_DLL_NAME_LENGTH - sizeof (WCHAR)))
	{
		if (ShowSnaps)
			DbgPrint
				("LdrResolveDllName - Unable To Locate %ws from %ws",
				DllName,
				SearchPath);
		RtlFreeAnsiString
			(PathFileName->Buffer);
		return 0;
	}
	// Construct PathFileName and FileName from
	// the results of the RtlDosSearchPath_U call.
	return 1;
} // LdrpResolveDllName

////////////////////////////////////////////////////////////////////////////////
HANDLE 
	LdrpCreateDllSection
		(PUNICODE_STRING ObjectName,
		HANDLE FileHandle,
		PUNICODE_STRING FileName,	// NOT USED
		LPCTSTR DllName,			// NOT USED
		HANDLE * SectionHandle)
{
	IO_STATUS_BLOCK IoStatusBlock;		// ebp-0xC
	OBJECT_ATTRIBUTES ObjectAttributes; // ebp-0x24
	if (0 == FileHandle)
	{
		ObjectAttributes.Length = sizeof (OBJECT_ATTRIBUTES);
		ObjectAttributes.RootDirectory = 0;
		ObjectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
		ObjectAttributes.ObjectName = ObjectName;
		ObjectAttributes.SecurityDescriptor = NULL;
		ObjectAttributes.SecurityQualityOfService = NULL;
		NTSTATUS rc =
			NtOpenFile
				(&FileHandle,
				SYNCHRONIZE | FILE_EXECUTE,
				&ObjectAttributes,
				&IoStatusBlock,
				FILE_SHARE_READ | FILE_SHARE_DELETE,
				FILE_SYNCHRONOUS_IO_NONALERT | FILE_NON_DIRECTORY_FILE);
		if (STATUS_SUCCESS != rc)
		{
			SectionHandle = NULL;
			return (HANDLE) rc;
		}
	}
	NtCreateSection
		(SectionHandle,
		SECTION_QUERY | SECTION_MAP_WRITE | SECTION_MAP_READ | SECTION_MAP_EXECUTE,
		NULL,		// ObjectAttributes
		0,			// SectionSize
		PAGE_EXECUTE,
		SEC_IMAGE,
		FileHandle);
	NtClose
		(FileHandle);
	if (FileHandle < 0)
		NtRaiseHardError
			(STATUS_INVALID_IMAGE_FORMAT);
	return SectionHandle;
} // LdrpCreateDllSection

////////////////////////////////////////////////////////////////////////////////
PMODULEITEM
	LdrpAllocateDataTableEntry
		(HMODULE hDll)
{
	PIMAGE_NT_HEADERS pNtHeaders =
		RtlImageNtHeader
			(hDll);
	PMODULEITEM pModuleItem = NULL;
	if (NULL == pNtHeaders)
		return (pModuleItem);
	pModuleItem =
		RtlAllocateHeap
			(GetProcessHeap(),
			(NtdllBaseTag + MEM_MAPPED) | HEAP_ZERO_MEMORY,
			sizeof (MODULEITEM));
	if (NULL == pModuleItem)
		return (pModuleItem);
	pModuleItem^.ImageBase = hDll;
	pModuleItem^.ImageSize = pNtHeaders^.OptionalHeader.SizeOfImage;
	pModuleItem^.TimeDateStamp = pNtHeaders^.FileHeader.TimeDateStamp;
	return pModuleItem;
} // LdrpAllocateDataTableEntry

////////////////////////////////////////////////////////////////////////////////
DWORD
	LdrpFetchAddressOfEntryPoint
		(HMODULE hDll)
{
	PIMAGE_NT_HEADERS pNtHeaders =
		RtlImageNtHeader
			(hDll);
	if (pNtHeaders^.OptionalHeader.AddressOfEntryPoint == 0)
		return (hDll);
	else
		return (hDll + pNtHeaders^.OptionalHeader.AddressOfEntryPoint);
} // LdrpFetchAddressOfEntryPoint

////////////////////////////////////////////////////////////////////////////////
void
	LdrpInsertMemoryTableEntry
		(PMODULEITEM pModuleItem)
{
	ULONG PEB = 0x7FFDF000;
	PMODULELISTHEADER pModuleList = PEB + 0xc;
	BYTE c =
		RtlUpcaseUnicodeChar
			(pModuleItem->FileName.Buffer[0]);
	c--;
	c &= 0x1F;
	// Update LdrpHashTable with pModuleItem->Unknown1 and pModuleItem->Unknown2 and vice versa.
	//
	// LdrpHashTable contains a maximum of 31 pairs of DWORDs corresponding to the characters:
	// A through Z, [, backslash, ], ^, _.  These DWORDs hold the addresses to pModuleItem->Unknown1
	// and pModuleItem->Unknown2 for the last pModuleItem beginning with the appropriate character.  
	// This effectively provides a quick lookup (hash) for LdrpCheckForLoadedDll when it is walking
	// the import descriptor for a loading DLL.
	pModuleItem->Unknown1 = *((PULONG) (LdrpHashTable + (c * sizeof (MODULELISTLINKS))));
	pModuleItem->Unknown2 = (LdrpHashTable + 4) + (c * sizeof (MODULELISTLINKS));
	*((PULONG) (LdrpHashTable + 4) + (c * sizeof (MODULELISTLINKS))) = pModuleItem->Unknown1;
	*((PULONG) ((LdrpHashTable + (c * sizeof (MODULELISTLINKS))) + 4)) = pModuleItem->Unknown2;
	// Attach the pModuleItem to end of the LoadOrder and MemOrder doubly-linked
	// lists fixing up the forward and backward links appropriately.
	pModuleItem->LoadOrder.Next = pModuleList->LoadOrder.Next;
	pModuleItem->LoadOrder.Prev = pModuleList->LoadOrder.Prev;
	(PMODULEITEM) (pModuleList->LoadOrder.Prev)->LoadOrder.Next = pModuleItem;
	pModuleList->LoadOrder.Prev = pModuleItem;
	pModuleItem->MemOrder.Prev = pModuleList->MemOrder.Prev;
	pModuleItem->MemOrder.Next = pModuleList->MemOrder.Next;
	(PMODULEITEM) (pModuleList->MemOrder.Next)->MemOrder.Next = pModuleItem->MemOrder.Next;
	pModuleList->MemOrder.Next = pModuleItem->MemOrder.Next;
} // LdrpInsertMemoryTableEntry

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpMapDll
		(LPCWSTR SearchPath,
		LPCWSTR DllName,
		PULONG ResolveDllReferences,
		ULONG StaticLink,
		NTSTATUS MapDllReturnCode,
		PMODULEITEM * pModuleItem)
{
	ULONG Work04;							// ebp-0x4
	PVOID pBaseAddress; 					// ebp-0x8
	HANDLE SectionHandle;					// ebp-0xC
	ULONG ViewSize; 						// ebp-0x10
	ULONG Response; 						// ebp-0x14
	ULONG FileHandle;						// ebp-0x18
	UNICODE_STRING PathFileName;			// ebp-0x38
	UNICODE_STRING FileName;				// ebp-0x40
	HARDERROR_REPONSE HardErrorResponse;	// ebp-0x48
	UNICODE_STRING ObjectName;				// ebp-0x58

	PVOID PEB = (PVOID) 0x7FFDF000;
	Work04 = 0;
	SectionHandle = 0;
	if (LdrpKnownDllObjectDirectory)
		if (!MapDllReturnCode)
		{
			if (DllName)
			{
				if ((wcschr (DllName, '\\') != 0)
						||
				   (wcschr (DllName, '/') != 0))
					;
				else
					SectionHandle =
						LdrpCheckForKnownDll
							(DllName
							&PathFileName,
							&FileName);
			}
			else
				SectionHandle =
					LdrpCheckForKnownDll
						(DllName
						&PathFileName,
						&FileName);
		}
	if (0 == SectionHandle)
	{
		ULONG ResolveDllName
			LdrpResolveDllName
				(SearchPath,
				DllName,
				&PathFileName,
				&FileName,
				&FileHandle);
		if (0 == ResolveDllName)
		{
			ResolveDllName = STATUS_DLL_NOT_FOUND;
			if (StaticLink)
			{
				// Set up parameters for NtRaiseHardError.
				NtRaiseHardError
					(...);
				if (!LdrpInLdrInit)
					LdrpFatalHardErrorCount++;
			}
			return ResolveDllName;
		}
		if (ShowSnaps)
			if (StaticLink)
				DbgPrint
					("LDR: Loading (%s) %wZ",
					"STATIC",
					PathFileName);
			else
				DbgPrint
					("LDR: Loading (%s) %wZ",
					"DYNAMIC",
					PathFileName);
		if (0 ==
			RtlDosPathNameToNtPathName
				(PathFileName.Buffer,
				&ObjectName,	// Format return: "\??\C:\PROJECTS\LOADLIBRARY\DEBUG\FORWARDED.DLL"
				0,
				0))
			return STATUS_OBJECT_PATH_SYNTAX_BAD;
		MapDllReturnCode =
			LdrpCreateDllSection
				(ObjectName,
				&FileHandle,
				&FileName,				// not used
				ResolveDllReferences,	// not used
				&SectionHandle);
		RtlFreeHeap
			(GetProcessHeap(),
			0,
			ObjectName.Buffer);
		if (0 > MapDllReturnCode)
		{
			RtlFreeAnsiString
				(PathFileName);
			RtlFreeAnsiString
				(FileName);
			return MapDllReturnCode;
		}
	}
	else
		Work04 = 1;
	pBaseAddress = 0;
	ViewSize = 0;
	ULONG SaveSubSystemData = PEB->SubSystemData;		// + 0x14
	PEB->SubSystemData = PathFileName.Buffer;
	// Dll load notification seen in debuggers 
	// will be produced by this call.
	MapDllReturnCode =
		NtMapViewOfSection
			(SectionHandle,
			GetCurrentProcess(),
			&pBaseAddress,
			0,				// ZeroBits
			0,				// CommitSize
			0,				// SectionOffset
			&ViewSize,
			ViewShare,
			0,				// AllocationType
			SECTION_MAP_READ);
	PEB->SubSystemData = SaveSubSystemData;
	if (STATUS_SUCCESS <= MapDllReturnCode)
	{
		(PIMAGE_NT_HEADERS) DllName =
			RtlImageNtHeader
				(pBaseAddress);
		if (NULL == DllName)
		{
			NtUnmapViewOfSection
				(GetCurrentProcess(),
				pBaseAddress);
			NtClose
				(SectionHandle);
			return STATUS_INVALID_IMAGE_FORMAT;
		}
		pModuleItem =
			LdrpAllocateDataTableEntry
				(pBaseAddress);
		if (NULL == pModuleItem)
		{
			NtUnmapViewOfSection
				(GetCurrentProcess(),
				pBaseAddress);
			NtClose
				(SectionHandle);
			return STATUS_NO_MEMORY;
		}
		pModuleItem->LoadCount = 0;
		__asm
		{
			MOV		AL,StaticLink
			...             
			NEG		AL                        
			SBB		EAX,EAX                   
			...             
			AND		EAX,STATIC_LINK                   
			MOVZX	EAX,AX                    
			MOV		pModuleItem->ModuleFlags,EAX 
		}
		pModuleItem->PathFileName.Length = PathFileName.Length;
		pModuleItem->PathFileName.MaximumLength = PathFileName.MaximumLength;
		pModuleItem->PathFileName.Buffer = PathFileName.Buffer;
		pModuleItem->FileName.Length = FileName.Length;
		pModuleItem->FileName.MaximumLength = FileName.MaximumLength;
		pModuleItem->FileName.Buffer = FileName.Buffer;
		pModuleItem->EntryPoint = 
			LdrpFetchAddressOfEntryPoint
				(pModuleItem->ImageBase);
		LdrpInsertMemoryTableEntry
			(pModuleItem);
		if (STATUS_IMAGE_MACHINE_TYPE_MISMATCH == MapDllReturnCode)
		{
			// Set up parameters for NtRaiseHardError.
			NtRaiseHardError
				(...);
			// Undo the mapping
			if (!LdrpInLdrInit)
				LdrpFatalHardErrorCount++;
			return STATUS_IMAGE_MACHINE_TYPE_MISMATCH;
		}
		if (pNtHeaders->FileHeader.Characteristics && IMAGE_FILE_LARGE_ADDRESS_AWARE == 0)
			pModuleItem->ModuleFlags |= IMAGE_DLL;		// 0x4
		if (pModuleItem->ModuleFlags && IMAGE_DLL != IMAGE_DLL)
			pModuleItem->EntryPoint = 0;
		if (MapDllReturnCode == STATUS_IMAGE_NOT_AT_BASE)
		{
			// Processing not covered in the article.
		}
		if ((LdrpNumberOfProcessors > 1)
				&&
		   (pModuleItem)
				&&
		   (pModuleItem->ModuleFlags & IMAGE_DLL))
			LdrpValidateImageForMp
				(pModuleItem);
	}
	NtClose
		(SectionHandle);
	return (MapDllReturnCode);
} // LdrpMapDll