////////////////////////////////////////////////////////////////////////////////
//
//	Russell Osterlund, May 2001
//
//	Pseudocode for:
//		LdrpCheckForLoadedDllHandle
//		LdrpNameToOrdinal
//		LdrpLoadImportModule
//		LdrpGetProcedureAddress
//		LdrpSnapThunk
//		LdrpSnapIAT
//		LdrpWalkImportDescriptor
//
////////////////////////////////////////////////////////////////////////////////
#include <WinNt.h>
#include "LdrpLoadDll.h"

////////////////////////////////////////////////////////////////////////////////
// Return 1, if found; 0, if not.
BYTE
	LdrpCheckForLoadedDllHandle
		(HMODULE hDll,
		PMODULEITEM *pModuleItem)	// OUTPUT
{
	if (LdrpLoadedDllHandleCache)
		if (((PMODULEITEM) LdrpLoadedDllHandleCache)->ImageBase == hDll)
		{
			*pModuleItem = (PMODULEITEM) LdrpLoadedDllHandleCache;
			return 1;
		}
	PVOID PEB = 0x7FFDF000;
	PMODULELISTLINKS pLoadOrderHeaderLink =
		((PMODULELISTHEADER) (PEB + 0xc))->LoadOrder;
	// If the list is empty, exit and return 0.
	PMODULEITEM pWork = 
		(PMODULEITEM) pLoadOrderHeaderLink->Next;
	while (pWork)
	{
		if ((pWork->MemOrder.Next)
				&&
			(pWork->ImageBase == hDll))
		{
			LdrpLoadedDllHandleCache = pWork;
			*pModuleItem = pWork;
			return 1;
		}
		pWork = pWork->LoadOrder.Next;
	}
	return 0;
} // LdrpCheckForLoadedDllHandle

////////////////////////////////////////////////////////////////////////////////
SHORT
	LdrpNameToOrdinal
		(LPCSTR ImportedAPIName,
		ULONG NumberOfFunctions,
		ULONG ImageBase,
		ULONG AddressOfNames,
		ULONG AddressOfNameOrdinals)
{
	ULONG Count &= 0;						// ebp-0x04
	if (NumberOfFunctions - 1 < 1)
		if (NumberOfFunctions - 1 < 0)
			return -1;
		else
			return *((PSHORT) AddressOfNameOrdinals + (NumberOfFunctions * sizeof (USHORT)));
	// Perform binary search on export table (located at AddressOfNames) 
	// comparing ImportedAPIName with the entries in the table.
	// If there is a match, return the ordinal number for the exported API; otherwise, return -1.
} // LdrpNameToOrdinal

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpLoadImportModule
		(LPCWSTR SearchPath,
		LPSTR ImportModule,			// used as work area by the routine
		ULONG ImageBase,
		PMODULE_ITEM *pModuleItem,
		PULONG CheckForLoadedDllReturnCode)
{
	PANSI_STRING Work;				// ebp-0x8
	PUNICODE_STRING TEBWorkString = TEB + 0xBF8;
	RtlInitAnsiString
		(Work,
		ImportModule);
	(NTSTATUS) ImportModule = 
		RtlAnsiStringToUnicodeString
			(TEBWorkString,
			Work,
			FALSE);			// AllocateDestinationString
	if ((NTSTATUS) ImportModule < 0) 
		return (NTSTATUS) ImportModule;
	UCHAR ReturnCode =
		LdrpCheckForLoadedDll
			(SearchPath,
			TEBWorkString,
			1,
			0,
			pModuleItem)
	if (ReturnCode)
	{
		// The import module has already been loaded.
		*CheckForLoadedDllReturnCode = 1;
		return (NTSTATUS) ImportModule;
	}
	else
	{
		// The import module has not already been loaded.
		*CheckForLoadedDllReturnCode = 0;
		NTSTATUS NtStatus =
			LdrpMapDll
				(SearchPath,
				TEBWorkString->Buffer,
				0,
				1,
				0,
				pModuleItem)
		if (STATUS_SUCCESS > NtStatus)
			return NtStatus;
		// Note! Note! - recursion - Note! Note!
		(NTSTATUS) ImportModule =
			LdrpWalkImportDescriptor
				(SearchPath,
				pModuleItem);
		if ((NTSTATUS) ImportModule < STATUS_SUCCESS)
		{
			// Remove the import module from InitOrder links.
		}
		return (NTSTATUS) ImportModule;
	}
} // LdrpLoadImportModule

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpGetProcedureAddress
		(ULONG ForwardedDLLImageBase,
		PUNICODE_STRING APIName,
		ULONG Ordinal,
		PULONG pIATEntry,
		PULONG pLdrpSnapThunkReturnValue)
{
	ULONG Unknown04;								// ebp-0x4
	PHINT_NAME_ENTRY pHintNameEntry;				// ebp-0x1c
	PIMAGE_EXPORT_DIRECTORY pImageExportDirectory;	// ebp-0x20
	ULONG IATEntry;									// ebp-0x24
	HINT_NAME_ENTRY WorkEntry;						// ebp-0x68
	PMODULEITEM pModuleItem;						// ebp-0x6c
	NTSTATUS NtStatus;								// ebp-0x70
	ULONG ImageExportDirectorySize;					// ebp-0x78
	ULONG LoadingItemImageBase;						// ebp-0x7c

	// This routine is wrapped by a __try/__except block.
	if (ShowSnaps)
		DbgPrint
			("LDR: LdrpGetProcedureAddress by ");
	pHintNameEntry = 0;
	if (NULL == APIName)
	{
		LoadingItemImageBase = 0;
		if (ShowSnaps)
			DbgPrint
				("ORDINAL - %lx",
				Ordinal);
		if (NULL == Ordinal)
			return STATUS_INVALID_PARAMETER;
		Ordinal |= 0x80000000;
		IATEntry = Ordinal;
	}
	else
	{
		if (ShowSnaps)
			DbgPrint
				("NAME - %s",
				APIName->Buffer);
		if (APIName->Length > 63)
		{
			pHintNameEntry =
				RtlAllocateHeap
					(GetCurrentProcess(),
					NtDllBaseTag + 0x180000,
					APIName->Length + 3);
			if (NULL == pHintNameEntry)
				return STATUS_INVALID_PARAMETER;
		}
		else
			pHintNameEntry = &WorkEntry;

		// copy APIName to HintNameEntry and null terminate HintNameEntry.

		LoadingItemImageBase = pHintNameEntry;
		IATEntry = 0;
	}
	if (LdrpInLdrInit)
		RtlEnterCriticalSection
			(PEB->CriticalSectionArea);		// +0xA0
	Unknown04 = 0;
	if (0 ==
			LdrpCheckForLoadedDllHandle
				(ForwardedDLLImageBase,
				&pModuleItem))
	{
		_local_unwind2
			(...);
		return STATUS_DLL_NOT_FOUND;
	}
	pImageExportDirectory =
		(PIMAGE_IMPORT_DESCRIPTOR) RtlImageDirectoryEntryToData
			(pModuleItem->ImageBase,
			1,
			IMAGE_DIRECTORY_ENTRY_EXPORT,
			&ImageExportDirectorySize);
	if (NULL == pImageImportDescriptor)
	{
		_local_unwind2
			(...);
		return STATUS_PROCEDURE_NOT_FOUND;
	}
	NtStatus =
		LdrpSnapThunk
			(pModuleItem->ImageBase,
			LoadingItemImageBase,
			IATEntry,
			IATEntry,
			pImageExportDirectory,
			ImageExportDirectorySize,
			0,
			0);
	if (0 != *(PBYTE) pLdrpSnapThunkReturnValue)
	{
		// processing related to GetProcAddress processing
		if (/* The last entry in the InitOrder has not been processed, 
			i.e. ModuleFlags & ENTRY_PROCESSED == 0 */)
			LdrpRunInitializeRoutines
				(0);
	}
	if (STATUS_SUCCESS >= NtStatus)
		*pIATEntry = IATEntry;
	Unknown04 |= 0xff;
	if (NULL != pHintNameEntry)
		if (pHintNameEntry->Name != &WorkEntry)
			RtlFreeHeap
				(GetCurrentProcess(),
				0,
				pHintNameEntry);
	if (LdrpInLdrInit)
		RtlLeaveCriticalSection
			(PEB->CriticalSectionArea);		// +0xA0
	return NtStatus;
} // LdrpGetProcedureAddress

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpSnapThunk
		(ULONG AdditionalItemImageBase,
		ULONG LoadingItemImageBase,
		PULONG pImportThunkEntry,
		PULONG pIATEntry,
		PIMAGE_EXPORT_DIRECTORY pImageExportDirectory,
		PULONG pExportDirectoryDataSize,
		ULONG Unknown20,
		LPSTR ImportedDLLName)
{
	ULONG Hint;											// ebp-0x4
	ANSI_STRING DllNameAnsi;							// ebp-0x8
	UNICODE_STRING DllNameUnicode;						// ebp-0x10
	NTSTATUS NtStatus = STATUS_SUCCESS;
	USHORT Ordinal;

	pImportThunkEntry = (*pImportThunkEntry) & 0x80000000;
	if (pImportThunkEntry)	// Import by ordinal?
	{
		// Note - This is not the same as performing a GetProcAddress by ordinal.
	}
	else
	{
		PHINT_NAME_ENTRY pHintNameEntry =
			(PHINT_NAME_ENTRY) (LoadingItemImageBase + *pImportThunkEntry);
		Hint = pHintNameEntry->Hint;
		if (pHintNameEntry->Hint > pImageExportDirectory->NumberOfNames)
			Ordinal =
				LdrpNameToOrdinal
					(pHintNameEntry->Name,
					pImageExportDirectory->NumberOfFunctions,
					AdditionalImageBase,
					pImageExportDirectory->AddressOfNames,
					pImageExportDirectory->AddressOfNameOrdinals);
		else
		{
			PCHAR ExportedName =
				(PCHAR) ((AdditionalItemImageBase + pImageExportDirectory->AddressOfNames) + 
				(Hint * sizeof (ULONG)));
			if (strcmp (pHintNameEntry->Name, ExportedName) == 0)
				Ordinal =
					AdditionalImageBase + 
					pImageExportDirectory->AddressOfNameOrdinals + 
					(Hint * sizeof (USHORT));
			else
			{
				__asm	mov	bl, (PUCHAR) pImportThunkEntry
				Ordinal =
					LdrpNameToOrdinal
						(pHintNameEntry->Name,
						pImageExportDirectory->NumberOfNames,
						AdditionalImageBase,
						pImageExportDirectory->AddressOfNames,
						pImageExportDirectory->AddressOfNameOrdinals);
			}
		}
	}
	if (Ordinal > pImageExportDirectory->NumberOfFunctions)
	{
		NtStatus = STATUS_ENTRYPOINT_NOT_FOUND;
		...
		*pIATEntry = 0xFFBADD11;
		return NtStatus;
	}
	ULONG EntryPoint = 
		AdditionalImageBase + 
		pImageExportDirectory->AddressOfFunctions + 
		(Ordinal * sizeof (ULONG));
	//
	// Update the IAT entry with the calculated entry-point:
	//   runtime binding
	//
	*pIATEntry = EntryPoint;
	//
	// See section 6.3.2 of PE specifications -- 
	//   Is the entrypoint of the API inside of the export directory?
	if ((EntryPoint > AdditionalImageBase + pImageExportDirectory)
				&&
		(EntryPoint < AdditionalImageBase + pImageExportDirectory + *pExportDirectoryDataSize))
	{
		// This API is forwarded.
		// Forwarded entries look like "NTDLL.RtlAllocateHeap"; grab the DLL name.
		PCHAR Period = 
			strchr 
				(AdditionalImageBase + 
					pImageExportDirectory->AddressOfNames + 
					(Ordinal * sizeof (ULONG)), 
				'.');
		DllNameAnsi.Length = 
			Period - 
			AdditionalImageBase + 
			pImageExportDirectory->AddressOfNames + 
			(Ordinal * sizeof (ULONG));
		DllNameAnsi.MaximumLength = 
			Period - 
			AdditionalImageBase + 
			pImageExportDirectory->AddressOfNames + 
			(Ordinal * sizeof (ULONG));
		DllNameAnsi.Buffer = 
			AdditionalImageBase + 
			pImageExportDirectory->AddressOfNames + 
			(Ordinal * sizeof (ULONG));
		if (STATUS_SUCCESS >= 
				RtlAnsiStringToUnicodeString 
					(DllNameUnicode, 
					DllNameAnsi, 
					TRUE))			// AllocateDestinationString
		{
			NtStatus =
				LdrpLoadDll
					(NULL,						// NULL Searchpath
					0,
					DllNameUnicode,				// Note - no ".DLL" extension
					pImageExportDirectory,
					0);							// forwarder processing
			RtlFreeAnsiString 
				(DllNameUnicode);
		}
		if (STATUS_SUCCESS > NtStatus)
		{
			NtStatus = STATUS_ENTRYPOINT_NOT_FOUND;
			...
			*pIATEntry = 0xFFBADD11;
			return NtStatus;
		}
		// Forwarded entries look like "NTDLL.RtlAllocateHeap"; now grab the API name.
		if (STATUS_SUCCESS ==
				RtlInitAnsiString
					(DllNameAnsi,
					DllNameAnsi.Length + 
						AdditionalImageBase + 
						pImageExportDirectory->AddressOfNames + 
						(Ordinal * sizeof (ULONG))))
		{
			if (DllNameAnsi.Buffer == '#')
				// Ordinal present rather than a string?
				if (STATUS_SUCCESS >
						RtlCharToInteger
							(DllNameAnsi.Buffer + 1,
							0,
							&AdditionalImageBase))
				{
					NtStatus = STATUS_ENTRYPOINT_NOT_FOUND;
					...
					*pIATEntry = 0xFFBADD11;
					return NtStatus;
				}
			if (STATUS_SUCCESS >
					LdrpGetProcedureAddress
						(pImageExportDirectory,
						DllNameAnsi,
						AdditionalImageBase,	// Can contain an ordinal.
						pIATEntry,				// this field gets plugged with the
												// forwarded entry-point for the API.
						0))
			{
				NtStatus = STATUS_ENTRYPOINT_NOT_FOUND;
				...
				*pIATEntry = 0xFFBADD11;
				return NtStatus;
			}
			NtStatus = STATUS_SUCCESS;
		}
		else
		{
		}
	}
	if (0 == EntryPoint)
	{
		NtStatus = STATUS_ENTRYPOINT_NOT_FOUND;
		...
		*pIATEntry = 0xFFBADD11;
		return NtStatus;
	}
	return NtStatus;
} // LdrpSnapThunk

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpSnapIAT
		(PMODULEITEM pAdditionalItem,
		PMODULEITEM pLoadingItem,
		PIMAGE_IMPORT_DESCRIPTOR pImageImportDescriptor,
		PBYTE pFlag)
{
	ULONG Unknown4;										// ebp-0x4
	PULONG pImageIATDirectory;							// ebp-0x1C
	PIMAGE_EXPORT_DIRECTORY pImageExportDirectory;		// ebp-0x20
	PULONG pIATEntry;									// ebp-0x28
	LPSTR ImportedDLLName;								// ebp-0x2c
	ULONG IATDirectoryDataSize;							// ebp-0x34
	NTSTATUS LdrpSnapThunkResult;						// ebp-0x3C
	ULONG ExportDirectoryDataSize;						// ebp-0x48
	PULONG pImportThunkEntry;							// ebp-0x4C
	ULONG IATFirstThunkEntry;							// ebp-0x50
	ULONG OldProtect;									// ebp-0x54
	ULONG SizeProtect;									// ebp-0x58
	ULONG LoadingImageBase;								// ebp-0x68

	// This routine is wrapped by a __try/__except block.
	pImageExportDirectory =
		(PIMAGE_EXPORT_DIRECTORY) RtlImageDirectoryEntryToData
			(pAdditionalItem->ImageBase,
			1,
			IMAGE_DIRECTORY_ENTRY_EXPORT,
			&ExportDirectoryDataSize);
	if (NULL == pImageExportDirectory)
		return STATUS_INVALID_IMAGE_FORMAT;
	LoadingImageBase = pLoadingItem->ImageBase;
	pImageIATDirectory =
		(PULONG) RtlImageDirectoryEntryToData
			(pLoadingItem->ImageBase,
			1,
			IMAGE_DIRECTORY_ENTRY_IAT,
			&IATDirectoryDataSize);
	if (NULL == pImageIATDirectory)
	{
		...
	}
	SizeProtect = IATDirectoryDataSize;
	ULONG rc =
		NtProtectVirtualMemory
			(GetCurrentProcess(),
			pImageIATDirectory,
			SizeProtect,
			PAGE_READWRITE,
			&OldProtect);
	if (STATUS_SUCCESS > rc)
		return rc;
	if (0 == *pFlag)
	{
		if (pImageImportDescriptor->FirstThunk)
		{
			pIATEntry = pLoadingItem->ImageBase + pImageImportDescriptor->FirstThunk;
			PIMAGE_NT_HEADERS32 pImageNtHeader =
				RtlImageNtHeader
					(pLoadingItem->ImageBase);
			if ((pImageImportDescriptor->OriginalFirstThunk < pImageNtHeader->OptionalHeader.SizeOfHeaders)
					||
				(pImageImportDescriptor >= pImageNtHeader->OptionalHeader.SizeOfImage))
				pImportThunkEntry = pIATEntry;
			else
				pImportThunkEntry = pLoadingItem->ImageBase + pImageImportDescriptor->FirstThunk;
			ImportedDLLName = (LPSTR) (pImageImportDescriptor->Name + pLoadingItem->ImageBase);
			while (*pImportThunkEntry)
			{
				Unknown4 = 1;
				LdrpSnapThunkResult =
					LdrpSnapThunk
						(pAdditionalItem->ImageBase,
						pLoadingItem->ImageBase,
						pImportThunkEntry,
						pIATEntry,
						pImageExportDirectory,
						&ExportDirectoryDataSize,
						1,
						ImportedDLLName);
				rc = LdrpSnapThunkResult;
				pImportThunkEntry += sizeof (ULONG);
				pIATEntry += sizeof (ULONG);
				Unknown4 |= 0xFF;
				if (STATUS_SUCCESS > LdrpSnapThunkResult)
					break;
			}
		}
	}
	else
	{
		ImportedDLLName = (LPSTR) (pImageImportDescriptor->Name + pLoadingItem->ImageBase);
		if (-1 != pImageImportDescriptor->ForwarderChain)	// there are forwarders
		{
			IATFirstThunkEntry = 
				pLoadingItem->ImageBase +
				pImageImportDescriptor->FirstThunk +
				(pImageImportDescriptor->ForwarderChain / 4);
			while (IATFirstThunkEntry)
			{
				Unknown4 &= 0;
				LdrpSnapThunkResult =
					LdrpSnapThunk
						(pAdditionalItem->ImageBase,
						pLoadingItem->ImageBase,
						pImageImportDescriptor->OriginalFirstThunk + 
							pLoadingItem->ImageBase + 
							(pImageImportDescriptor->ForwarderChain / 4),
						IATFirstThunkEntry,
						pImageExportDirectory,
						&ExportDirectoryDataSize,
						1,
						ImportedDLLName);
				rc = LdrpSnapThunkResult;
				IATFirstThunkEntry += sizeof (ULONG);
				pIATEntry = IATFirstThunkEntry;
				Unknown4 |= 0xFF;
				if (STATUS_SUCCESS > LdrpSnapThunkResult)
					break;
			}
		}
	}
	NtProtectVirtualMemory
		(GetCurrentProcess(),
		pImageIATDirectory,
		SizeProtect,
		OldProtect,
		&OldProtect);
	NtFlushInstructionCache
		(GetCurrentProcess(),
		pImageIATDirectory,
		IATDirectoryDataSize);
	return rc;
} // LdrpSnapIAT

////////////////////////////////////////////////////////////////////////////////
NTSTATUS
	LdrpWalkImportDescriptor
		(LPCWSTR SearchPath,
		PMODULEITEM pModuleItem)
{
	NTSTATUS ReturnCode;											// ebp-0x4
	PMODULEITEM pWorkItem;											// ebp-0x8
	PVOID PEB = 0x7FFDF000;											// ebp-0x10
	ULONG Unknown14;												// ebp-0x14
	ULONG ImageIATDirectorySize;									// ebp-0x18
	ULONG ImportDirectoryDataSize;									// ebp-0x1C
	ULONG BoundImportDirectoryDataSize;								// ebp-0x20
	PIMAGE_BOUND_IMPORT_DESCRIPTOR pImageBoundImportDescriptor;		// ebp-0x24
	PIMAGE_BOUND_IMPORT_DESCRIPTOR pImageBoundImportDescriptorNext;	// ebp-0x28
	LPSTR ImportDll;												// ebp-0x2C
	PIMAGE_BOUND_IMPORT_DESCRIPTOR pImageBoundImportDescriptorNext;	// ebp-0x28
	PULONG pImageIATDirectory;										// ebp-0x30
	pImageBoundImportDescriptor =
		(PIMAGE_BOUND_IMPORT_DESCRIPTOR) RtlImageDirectoryEntryToData
			(pModuleItem->ImageBase,
			1,
			IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT,
			&BoundImportDirectoryDataSize);
	PIMAGE_IMPORT_DESCRIPTOR pImageImportDescriptor =
		(PIMAGE_IMPORT_DESCRIPTOR) RtlImageDirectoryEntryToData
			(pModuleItem->ImageBase,
			1,
			IMAGE_DIRECTORY_ENTRY_IMPORT,
			&ImportDirectoryDataSize);
	if (pImageBoundImportDescriptor)
	{
		pImageBoundImportDescriptorNext = pImageBoundImportDescriptor;
		while ((pImageBoundImportDescriptorNext)
					&&
				(pImageBoundImportDescriptorNext->OffsetModuleName))
		{
			ImportDll =
				(LPSTR) (pImageBoundImportDescriptorNext + pImageBoundImportDescriptorNext->OffsetModuleName);
			if (ShowSnaps)
				DbgPrint
					("LDR: %wZ bound to %s",
					pModuleItem->FileName,
					ImportDll);
			ReturnCode =
				LdrpLoadImportModule
					(SearchPath,
					ImportDll,
					pModuleItem->ImageBase,
					&pWorkItem,
					&ReturnCode);
			if (STATUS_SUCCESS > ReturnCode)
				return ReturnCode;
			if (STATUS_SUCCESS == ReturnCode)
			{
				// Insert pWorkItem into the ModuleList's InitOrder chain (at the end)
				// and update the links in pWorkItem->InitOrder.
			}
			if (pImageBoundImportDescriptorNext->TimeDateStamp == pWorkItem->TimeDateStamp)
			{
				if (IMAGE_NOT_AT_BASE & pWorkItem->ModuleFlags)
				{
					if (ShowSnaps)
						DbgPrint
							("LDR: %wZ has stale binding to %s",
							pModuleItem->FileName,
							ImportDll);
					pModuleItem = 1;
					Unknown14 &= 0;
					pImageBoundImportDescriptorNext += sizeof (IMAGE_BOUND_IMPORT_DESCRIPTOR);
					// 0x77F8C257
				}
				else
				{
					if (ShowSnaps)
						DbgPrint
							("LDR: %wZ has correct binding to %s",
							pModuleItem->FileName,
							ImportDll);
					pModuleItem = 0;
				}
			}
			else
			{
				if (ShowSnaps)
					DbgPrint
						("LDR: %wZ has stale binding to %s",
						pModuleItem->FileName,
						ImportDll);
				pModuleItem = 1;
			}
			if (pModuleItem)
			{
				// Perform normal "on-the-fly" binding.
				// See the processing where the normal import descriptor is used.
			}
			Unknown14 &= 0;
			pImageBoundImportDescriptorNext += sizeof (IMAGE_BOUND_IMPORT_DESCRIPTOR);
		}
	}
	else
	if (pImageImportDescriptor)
	{
		while ((pImageImportDescriptor->Name)
					&&
			   (pImageImportDescriptor->FirstThunk))
		{
			ImportDll =
				(LPSTR) (pWorkItem->ImageBase + pImageImportDescriptor->Name);
			if (ShowSnaps)
				DbgPrint
					("LDR: %s used by %wZ",
					ImportDll,
					pModuleItem->FileName);
			ReturnCode =
				LdrpLoadImportModule
					(SearchPath,
					ImportDll,
					pModuleItem->ImageBase,
					&pWorkItem,
					&ReturnCode);
			if (ReturnCode < STATUS_SUCCESS)
				return ReturnCode;
			if (ShowSnaps)
				DbgPrint
					("LDR: Snapping imports for %wZ from %s",
					pModuleItem->FileName,
					ImportDll);
			*((PBYTE) pModuleItem) &= 0;
			if ((NULL == pImageImportDescriptor)
					||
				(0 == pImageImportDescriptor->TimeDateStamp)	// DLLs not bound will have zero here.
																// See Sec. 6.4.1 of PE specifications.
					||
				(pWorkItem->TimeDateStamp)
					||
				(pWorkItem->ModuleFlags & IMAGE_NOT_AT_BASE))
			{
				if (STATUS_SUCCESS == ReturnCode)
				{
					// Insert pWorkItem into the ModuleList's InitOrder chain (at the end)
					// and update the links in pWorkItem->InitOrder.
				}
				if (0 > 
					LdrpSnapIAT
						(pWorkItem,
						pModuleItem,
						pImageImportDescriptor,
						(PBYTE) pModuleItem))
						return ReturnCode;
				ReturnCode = 1;
			}
			else
			{
				if (ShowSnaps)
				{
					DbgPrint
						("LDR: Snap bypass %s from %wZ",
						ImportDll,
						pModuleItem->FileName);
				}
				*pModuleItem = 1;
			}
			// Look at the next entry in the import descriptor.
			pImageImportDescriptor += sizeof (IMAGE_IMPORT_DESCRIPTOR);
		}
	}
	if (*((PBYTE) (PEB + 0x69)) & 0x80)
	{
		// terra incognita
		ImageIATDirectorySize =
			RtlImageDirectoryEntryToData
				(pModuleItem->ImageBase,
				1,
				IMAGE_DIRECTORY_ENTRY_IAT,
				&pImageIATDirectory);
		pModuleItem = pImageIATDirectory;
		if (ImageIATDirectorySize)
		{
			if (STATUS_SUCCESS ==
				NtProtectVirtualMemory
					(...))
			{
				RtlAllocateHeap
					(...);
				LdrpDefineDllTag
					(...);
				NtProtectVirtualMemory
					(...);
			}
			else
				DbgPrint
					("LDR: Unable to unprotect IAT to enable tagging by DLL.");
		}
	}
	if (*(PBYTE)(PEB + 0x6B) & 0x02)
		// terra incognita
		LdrpDphInitializeTargetDll
			(pModuleItem);
	return STATUS_SUCCESS;
} // LdrpWalkImportDescriptor